import itertools
import yter


def test_call():
    assert tuple(yter.call((0, 1, lambda: 2))) == (0, 1, 2)
    assert tuple(yter.call(())) == ()


def test_percent():
    assert tuple(yter.percent((0, 1, 2), 0.5)) == (0, 2)

    numbers = tuple(range(100))
    assert tuple(yter.percent(numbers, 1)) == numbers
    assert tuple(yter.percent(numbers, 1.1)) == numbers
    assert tuple(yter.percent(numbers, 5.5)) == numbers

    assert len(tuple(yter.percent(numbers, 0))) == 0
    assert len(tuple(yter.percent(numbers, 0.0))) == 0
    assert len(tuple(yter.percent(numbers, 0.01))) == 1
    assert len(tuple(yter.percent(numbers, 0.11))) == 11
    assert len(tuple(yter.percent(numbers, .47))) == 47


def test_flat():
    assert tuple(yter.flat(((0, 1), (2,)))) == (0, 1, 2)


def test_chunk():
    assert tuple(map(tuple, yter.chunk((0, 1, 2), 2))) == ((0, 1), (2,))
    assert tuple(map(tuple, yter.chunk(itertools.chain((0, 1, 2)), 2))) == ((0, 1), (2,))


def test_key():
    assert tuple(yter.key(("one", "two", "three", "four"), len)) == (
        (3, "one"), (3, "two"), (5, "three"), (4, "four"))


def test_unique():
    assert tuple(yter.unique((1, 1, 2, 3, 1))) == (1, 2, 3)
    assert tuple(yter.unique((1, 1, 2, 3, 1), key=lambda k: 0)) == (1,)


def test_duplicates():
    assert tuple(yter.duplicates((1, 1, 2, 3, 1))) == (1, 1)
    assert tuple(yter.duplicates((1, 1, 2, 3, 1), key=lambda k: 0)) == (1, 2, 3, 1)




import pytest
import yter


def test_formatter():
    assert yter.formatter("val")(None) == "val"
    assert yter.formatter("{0:3d}")(22) == " 22"
    assert yter.formatter("{it}", it=True)(False) == "True"
    with pytest.raises(IndexError):
        yter.formatter("{1}")("one")



def test_numeric():
    assert yter.numeric("one2three") == ("one", 2, "three")
    assert yter.numeric("one-2three") == ("one", -2, "three")
    assert yter.numeric("1two3") == (1, "two", 3)
    assert yter.numeric("88") == (88,)
    assert yter.numeric("eight-eight") == ("eight-eight",)
    assert yter.numeric("eight-") == ("eight-",)
    assert yter.numeric("-") == ("-",)
    assert yter.numeric("") == ()


def test_getter():
    data = {"name": {"first": "Peter", "last": "Shinners"}, "id": 1234}
    key = yter.getter["name"]["first"]
    assert key(data) == "Peter"
    assert yter.getter.real(1.2) == 1.2
    assert yter.getter("nop") == "nop"
    assert yter.getter.waffle(data) == None
    assert yter.getter["name", "id"](data) == (data["name"], 1234)

    types = [bool, int, float, str]
    results = list(map(yter.getter._(12), types))
    assert results == [True, 12, 12.0, "12"]

    getter = yter.getter.split._("@", 2)
    assert getter("one@two@three@four") == ["one", "two", "three@four"]



"""Functions that make useful keys"""

import re
import operator


def formatter(fmt, **kwargs):
    """Create a function that formats given values into strings.

    Additional keywoard arguments can be given that will also be passed to
    the format function.

    The values passed to the formatter will be given as the first argument
    to format, which is referenecd as {0}.

    """
    def formatted(val):
        return fmt.format(val, **kwargs)
    return formatted


_numberRegex = None


def numeric(value):
    """Split a string into string and integer sections.

    A string will be a tuple containing sections that are strings
    and integers. The numeric parts of the string will be converted
    into integers.

    This is a convenient way to sort string containing numbers that
    are not padded.

    Negative numbers will also be converted if preceded by a single
    minus sign.

    """
    global _numberRegex

    if not value:
        return ()

    if not _numberRegex:
        _numberRegex = re.compile("(-?\d+)")
    parts = _numberRegex.split(value)
    # Values that start with digits will get a leading empty string
    if value[0].isdigit():
        parts = parts[1:]
    if value[-1].isdigit():
        parts.pop()

    converted = tuple(int(p) if p[-1].isdigit() else p
                for p in parts)
    return converted


class _GetterImpl(object):
    """Shorthand for the attrgetter, itemgetter, and methodcaller operators.

    The same results can be achieved by using `operator.attrgetter` and
    `operator.itemgetter`, but this is more concise and can also be used
    to lookup nested data.

    By using the special syntax, `getter._(args)` you can also create
    a callable. This is similar to the `operator.methodcaller` but it
    doesn't use a method name. The underscore attribute can be called
    to create a callable lookup, which can still be chained with item
    and attributes.

    If an attribute or item is not found this will result in None
    instead of an exception, which is more useful for key functions.

    You can lookup multiple items by passing multiple values to the
    index operator.

    Looking up attributes can only be done with a single value.

    >>> data = {"name": {"first": "Peter", "last": "Shinners"}, "id": 1234}
    >>> key = yter.getter["name"]["first"]
    >>> print(key(data))
    "Peter"
    >>> print(yter.getter.real(1.2))
    1.2
    >>> data = [bool, int, float, str]
    >>> print [yter.getter._("12") for d in data]
    [True, 12, 12.0, '12']

    """
    def __init__(self, getters):
        self.__getters = getters

    def __getitem__(self, vals):
        if isinstance(vals, tuple):
            getter = operator.itemgetter(*vals)
        else:
            getter = operator.itemgetter(vals)
        key = type(self)(self.__getters + (getter,))
        return key

    def __getattr__(self, name):
        getter = operator.attrgetter(name)
        key = type(self)(self.__getters + (getter,))
        return key

    @property
    def _(self):
        def _makecaller(*args, **kwargs):
            getter = lambda val: val(*args, **kwargs)
            key = type(self)(self.__getters + (getter,))
            return key
        return _makecaller

    def __call__(self, value):
        try:
            for getter in self.__getters:
                value = getter(value)
        except KeyError:
            return None
        except AttributeError:
            return None
        return value


getter = _GetterImpl(())



import itertools
import yter


def test_yany():
    assert yter.yany((0, 1, 2)) == 1
    assert yter.yany(itertools.chain(("a", 1, [2]))) == "a"
    assert yter.yany(iter(("", 0, []))) == []

    assert yter.yany((), empty=5) == 5
    assert yter.yany(iter(()), empty=5) == 5
    assert yter.yany((2,), empty=5) == 2

    assert yter.yany(("a", "aa", "aaa"), key=lambda x: len(x) - 1) == "aa"



def test_yall():
    assert yter.yall((0, 1, 2)) == 0
    assert yter.yall(itertools.chain(("a", 0, [2]))) == 0
    assert yter.yall(iter(("", 1, []))) == ""

    assert yter.yall((), empty=5) == 5
    assert yter.yall(iter(()), empty=5) == 5
    assert yter.yall((2,), empty=5) == 2

    assert yter.yall(("a", "aa", "aaa"), key=lambda x: len(x) - 1) == "a"


def test_first():
    assert yter.first((0, 1, 2)) == 0
    assert yter.first(itertools.chain((0, 1, 2))) == 0
    assert yter.first(iter(("", 1, []))) == ""

    fs = frozenset(("a", "b", "c"))
    assert yter.first(fs) in fs

    assert yter.first((), empty=5) == 5
    assert yter.first(iter(()), empty=5) == 5
    assert yter.first((2,), empty=5) == 2


def test_last():
    assert yter.last((0, 1, 2)) == 2
    assert yter.last(itertools.chain((0, 1, 2))) == 2
    assert yter.last(iter(("", 1, []))) == []

    fs = frozenset(("a", "b", "c"))
    assert yter.last(fs) in fs

    assert yter.last((), empty=5) == 5
    assert yter.last(iter(()), empty=5) == 5
    assert yter.last((2,), empty=5) == 2


def test_head():
    assert yter.head((0, 1, 2), 2) == [0, 1]
    assert yter.head((0, 1, 2), 5) == [0, 1, 2]
    assert yter.head(itertools.chain((0, 1, 2)), 1) == [0]
    assert yter.head(iter(("", 1, [])), 0) == []


def test_tail():
    assert yter.tail((0, 1, 2), 2) == [1, 2]
    assert yter.tail((0, 1, 2), 5) == [0, 1, 2]
    assert yter.tail(itertools.chain((0, 1, 2)), 1) == [2]
    assert yter.tail(iter(("", 1, [])), 0) == []


def test_ylen():
    assert yter.ylen((0, 1, 2)) == 3
    assert yter.ylen(itertools.chain((0, 1, 2))) == 3
    assert yter.ylen(iter(("", 1, []))) == 3
    y = iter("abc")
    assert yter.ylen(y) == 3
    assert yter.ylen(y) == 0

def test_minmax():
    assert yter.minmax((0, 1, 2)) == (0, 2)
    assert yter.minmax((2, 0, 1)) == (0, 2)
    assert yter.minmax(iter((0, 2, 1))) == (0, 2)
    assert yter.minmax(itertools.chain((1, 2, 0))) == (0, 2)

    assert yter.minmax((), empty=5) == (5, 5)
    assert yter.minmax(iter(()), empty=5) == (5, 5)

    # Test stability, equal vals preserve order
    class e(object):
        def __lt__(self, other):
            return False
        def __ge__(self, other):
            return True
    e1 = e()
    e2 = e()
    e3 = e()
    assert yter.minmax((e1, e3, e2)) == (e1, e2)

    yter.minmax(("one", "two", "three", "four"), key=len) == ("one", "three")


def test_minmedmax():
    assert yter.minmedmax((0, 1, 2)) == (0, 1, 2)
    assert yter.minmedmax(iter((0, 2, 1))) == (0, 1, 2)
    assert yter.minmedmax(itertools.chain((1, 2, 0))) == (0, 1, 2)

    assert yter.minmedmax((4, 7, 2, 1)) == (1, 4, 7)
    assert yter.minmedmax((4, 7, 2, 3, 1)) == (1, 3, 7)

    assert yter.minmedmax((), empty=5) == (5, 5, 5)
    assert yter.minmedmax(iter(()), empty=5) == (5, 5, 5)

    # Test stability, equal vals preserve order
    class e(object):
        def __lt__(self, other):
            return False
        def __ge__(self, other):
            return True
    e1 = e()
    e2 = e()
    e3 = e()
    assert yter.minmedmax((e1, e3, e2)) == (e1, e3, e2)

    yter.minmedmax(("one", "two", "three", "four"), key=len) == ("one", "two", "three")


def test_isiter():
    assert yter.isiter((1,))
    assert yter.isiter({})
    assert yter.isiter([1])
    assert yter.isiter(range(1))

    assert not yter.isiter(None)
    assert not yter.isiter(9)
    assert not yter.isiter("")
    assert not yter.isiter("a")

    assert yter.isiter(itertools.chain((1,)))


def test_sequence():
    l = [1, 2]
    assert yter.sequence(l) is l

    il = iter(l)
    assert yter.sequence(il) is not il

    d = {1: 1, 2: 2}
    assert yter.sequence(d) is d

    c = itertools.chain((1,))
    assert yter.sequence(c) is not c

    _xrange = __builtins__.get("xrange", __builtins__["range"])
    x = _xrange(3)
    assert yter.sequence(x) is x

    f = open(__file__, "rb")
    assert yter.sequence(f) is not f


def test_repeat():
    a = [1, 2, 3]
    assert yter.repeat(a) is a
    b = dict.fromkeys(a)
    assert yter.repeat(b) is b
    c = itertools.chain(a)
    d = yter.repeat(c)
    assert c is not d
    first = next(iter(d))
    assert list(d) == list(d)
    assert yter.repeat(d) is d



"""Clever, quick iterators that make your smile whiter."""

from ._fun import *
from ._ytr import *
from ._key import *

__all__ = [
    "yany",
    "yall",
    "ylen",
    "last",
    "head",
    "tail",
    "finish",
    "minmax",
    "minmedmax",
    "isiter",
    "contain",
    "repeat",
    "sequence",
    "call",
    "percent",
    "flat",
    "chunk",
    "key",
    "unique",
    "duplicates",
    "formatter",
    "numeric",
    "getter",
]



"""Iterators that wrap iterators for yter"""

import itertools


def call(y, *args, **kwargs):
    """Iterator that works with mixed callable types.

    Iterate over all the values from the input iterator. If one of the values
    is a callable object it will be called and its return will be the value
    instead.

    The `args` and `kwargs` arguments will be passed to any callable values
    in the iterator.

    """
    vals = []
    _callable = callable  # Local cache
    for val in y:
        if _callable(val):
            val = val(*args, **kwargs)
        yield val


def percent(y, percent):
    """Iterator that skips a percentage of values.

    The `percent` is a floating point value between 0.0 and 1.0. If the value
    is larger than 1.0 it will be the same as 1.0. If the value is less than
    or equal to 0.0 no values will be iterated.

    As long as the `percent` is greater than 0.0 the first value
    will always be iterated.

    """
    if percent <= 0.0:
        return iter(())
    elif percent >= 1.0:
        return y

    def genpercent(y, percent):
        num, denom = float(percent).as_integer_ratio()
        count = denom
        for val in y:
            if count >= denom:
                yield val
                count += num
                count -= denom
            else:
                count += num
    return genpercent(y, percent)


def flat(y):
    """Iterator of values from a iterable of iterators.

    This removes a level of nesting. When given a list of lists this will
    iterate the values from those children lists.

    This is the same as `itertools.chain.from_iterable` with a more memorable
    name.

    This will invert the results of the `chunk` iterator.
    """
    return itertools.chain.from_iterable(y)


def chunk(y, size):
    """Iterator of lists with a fixed size from iterable.

    Group the values from the iterable into lists up to a given size.
    The final list generated can be smaller than the given size.

    """
    y = iter(y)
    _len = len
    while 1:
        values = list(itertools.islice(y, size))
        l = _len(values)
        yield values
        if l < size:
            break


def key(y, key):
    """Iterator of pairs of key result and original values

    Each value will be a tuple with the return result of the `key`
    function and the actual value itself. The `key` function
    will be called with each value of the input iterator.

    This allows passing the generator to functions like `min` and `sorted`,
    but you are able to see the value and the result of the key argument.

    """
    for val in y:
        result = key(val)
        yield (result, val)


def unique(y, key=None):
    """Iterate only the unique values

    Only the first time a value is encountered it will be iterated.
    After that, values that are the same will be ignored.

    If the key is given, then it will be called on each value before
    comparing for uniqueness.

    """
    seen = set()
    seenadd = seen.add
    if key:
        for val in y:
            keyed = key(val)
            if keyed not in seen:
                seenadd(keyed)
                yield val
    else:
        for val in y:
            if val not in seen:
                seenadd(val)
                yield val


def duplicates(y, key=None):
    """Iterate only the duplicated values

    Will iterate a value for each time it is duplicated. If the iterator
    has three of the same value this will iterate the final two of them.

    If the key is given, then it will be called on each value before
    comparing for duplicates.

    """
    seen = set()
    seenadd = seen.add
    if key:
        for val in y:
            keyed = key(val)
            if keyed not in seen:
                seenadd(keyed)
            else:
                yield val
    else:
        for val in y:
            if val not in seen:
                seenadd(val)
            else:
                yield val




import setuptools
import re

readme = open("README.md").read()


setuptools.setup(
    name='yter',
    license='MIT',
    keywords='iterator itertools',
    description='Clever, quick iterators that make your smile whiter',

    author='Peter Shinners',
    author_email='pete@shinners.org',
    url='https://gitlab.com/shredwheat/yter',

    version=re.search(r"Version ([\d.]+)", readme).group(1),
    long_description=re.sub(r"\[(`.*?`)\]\(#.+?\)", r"\1", readme),

    packages=["yter"],

    zip_safe=True,
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
    ],   # https://pypi.python.org/pypi?%3Aaction=list_classifiers
)



"""Functions for getting data from iterators for yter"""

import sys
import bisect
import itertools
import collections


if sys.version_info[0] < 3:
    _stringTypes = (basestring,)
else:
    _stringTypes = (str, bytes)


def yany(y, key=None, empty=False):
    """Extended version of the builtin any, test if any values are true.

    Unlike the builtin `any` method, this will return the last value iterated
    before ending. This will short circuit or exit early when a correct answer
    is found.

    A simple explanation is that this will be the first true value or the
    final false value of the interator. This is the same as calling the
    logical or operator on all the iterated values.

    The `key` is an optional function that will be called on each value in
    iterator. The return of that function will be used to test if the value is
    true or false.

    If the  iterator is empty this will return the empty argument. This
    defaults to False, the same as the built in `any`.

    """
    val = empty
    if key:
        for val in y:
            if key(val):
                return val
    else:
        for val in y:
            if val:
                return val
    return val


def yall(y, key=None, empty=True):
    """Extended version of the builtin all, test if all values are true.

    Unlike the builtin `all` method, this will return the last value iterated
    before ending. This will short circuit or exit early when a correct answer
    is found.

    A simple explanation is that this will be the final true value
    or the first false value of the interator. This is the same as calling the
    logical and operator on all the iterated values.

    The `key` is an optional function that will be called on each value in
    iterator. The return of that function will be used to test if the value is
    true or false.

    If the iterable is empty this will return the empty argument. This
    defaults to True, the same as the built in `all`.

    """
    val = empty
    if key:
        for val in y:
            if not key(val):
                return val
    else:
        for val in y:
            if not val:
                return val
    return val


def first(y, empty=None):
    """Get the final value from an iterator.

    This will get the first value from an iterable object. This is an
    improvement over the builtin `next` because it works any iterable
    object, not just an iterator. If given an iterator it will be
    advanced by a single value.

    If the iterable contains no values then the `empty` argument is returned.

    """
    return next(iter(y), empty)


def last(y, empty=None):
    """Get the final value from an iterator.

    This will get the final value from an iterable object. If the iterable
    object can be `reversed` then that will be used to collect only the
    first value.

    If the iterable contains no values then the `empty` argument is returned.

    """
    try:
        ry = reversed(y)
    except TypeError:
        buf = collections.deque(y, maxlen=1)
        return buf[0] if buf else empty
    else:
        return next(ry, empty)


def head(y, count):
    """Get the first values from an iterator.

    This will be a list of values no larger than `count`. This will
    always advance the iterable object by the given count.

    """
    return list(itertools.islice(y, count))


def tail(y, count):
    """Get the last values from an iterator.

    This will be a list of values no larger than `count`.

    If the iterable object can be `reversed` then a more
    efficient algorithm is used. Otherwise this will
    always finish the iterable object.

    """
    try:
        ry = reversed(y)
    except TypeError:
        return list(collections.deque(y,  maxlen=count))
    else:
        items = list(itertools.islice(ry, count))
        items.reverse()
        return items


def ylen(y):
    """Complete an iterator and get number of iterations.

    Get all the values from an iterator and return the number of values
    it contained. An empty iterable will return 0.

    The values from the iterator are discarded.

    """
    span = l = 4096
    count = 0
    preroll = span * 3 // 2
    buf = [0] * (preroll + span)
    while l == span:
        buf[preroll:] = (itertools.islice(y, span))
        l = len(buf) - preroll
        count += l
    return count


def finish(y):
    """Complete an iterator and get number of iterations (deprecated).

    Get all the values from an iterator and return the number of values
    it contained. An empty iterable will return 0.

    The values from the iterator are discarded.

    This function has been renamed and is deprecated, see `ylen`.

    """
    return ylen(y)


def minmax(y, key=None, empty=None):
    """Find the minimum and maximum values from an iterable.

    This will always return a tuple of two values. If the iterable contains
    no values it the tuple will contain two values of the `empty` argument.

    The minimum and maximum preserve order. So the first value that compares
    equal will be considered the minimum and the last equal value is
    considered the maximum. If you sorted the iterable, this is the same as
    the first and last values from that list.

    The `key` is an optional function that will be called on each value in
    iterator. The return of that function will be used to sort the values
    from the iterator.

    """
    y = iter(y)
    try:
        minVal = maxVal = next(y)
    except StopIteration:
        return (empty, empty)

    if key:
        minKey = maxKey = key(minVal)
    else:
        minKey = maxKey = minVal

    for curVal in y:
        if key:
            curKey = key(curVal)
        else:
            curKey = curVal

        if curKey < minKey:
            minVal = curVal
            minKey = curKey
        elif curKey >= maxKey:
            maxVal = curVal
            maxKey = curKey
    return (minVal, maxVal)


def minmedmax(y, key=None, empty=None):
    """Find the minimum, median, and maximum values from an iterable.

    This will always return a tuple of three values. If the iterable contains
    no values it the tuple will contain three values of the `empty` argument.

    The minimum and maximum preserve order. So the first value that compares
    equal will be considered the minimum and the last equal value is
    considered the maximum. If you sorted the iterable, this is the same as
    the first and last values from that list.

    Computing the median requires storing half of the iterated values in
    memory. This only keeps the minimum required data, but be aware of
    this overhead when dealing with large iterators.

    The `key` is an optional function that will be called on each value in
    iterator. The return of that function will be used to sort the values
    from the iterator.

    """
    y = iter(y)
    try:
        minVal = maxVal = next(y)
    except StopIteration:
        return (empty, empty, empty)

    if key:
        minKey = maxKey = key(minVal)
    else:
        minKey = maxKey = minVal

    count = 1
    medianVals = [(minKey, minVal)]

    for curVal in y:
        if key:
            curKey = key(curVal)
        else:
            curKey = curVal

        if curKey < minKey:
            minVal = curVal
            minKey = curKey
        elif curKey >= maxKey:
            maxVal = curVal
            maxKey = curKey

        medVal = (curKey, curVal)
        bisect.insort_right(medianVals, medVal)
        count += 1
        if count % 2:
            medianVals.pop()

    medVal = medianVals[-1][1]
    return (minVal, medVal, maxVal)


def isiter(y, ignore=_stringTypes):
    """Test if an object is iterable, but not a string type.

    Test if an object is an iterator or is iterable itself. By default this
    does not return True for string objects.

    The `ignore` argument defaults to a list of string types that are not
    considered iterable. This can be used to also exclude things like
    dictionaries or named tuples. It will be used as an argument
    to `isinstance`. If ignore is set to None, it will be ignored.

    """
    if ignore and isinstance(y, ignore):
        return False
    try:
        iter(y)
        return True
    except TypeError:
        return False


# Deprecated, drop at some point
def contain(y):
    """Copy an iterator into a list if it is not already a sequence.

    If the value is already a container, like a list or tuple, then return
    the original value. Otherwise the value must be an iterator and it will
    be copied into a sequence.

    This function has been renamed and is deprecated, see `sequence`.

    """
    if iter(y) is not y:
        return y
    return list(y)


def sequence(y):
    """Efficient copy of non sequence data.

    If the value is already a container, like a list or tuple, then return
    the original value. Otherwise the value must be an iterator that will
    be copied into a new list.

    This allows an efficient way to repeat an iterator, without
    copying the data from a container.

    This is slightly more efficient than `repeat` because it
    copies all the data ahead of time. This is

    """
    if iter(y) is not y:
        return y
    return list(y)


def repeat(y):
    """Efficient lazy copy of non sequence data.

    If the value is already a sequence, that will be returned
    with no changes. Otherwise this will wrap the results in an
    iterator that lazily copies all values for repeated use.

    This allows efficient use of containers when no copy is needed
    to iterate the data multiple times.

    This is slightly less efficient than `sequence`, but if
    only iterating a portion of the data, this will not
    require a full copy.

    """
    yiter = iter(y)
    if yiter is not y:
        # Object does not need a separate iterator
        return y
    return _Repeat(y)


class _Repeat(object):
    """Iterator that can be reused any number of times"""
    def __init__(self, it):
        self.__tee1, self.__tee2 = itertools.tee(it)
    def __iter__(self):
        return self.__tee2.__copy__()
